"""
This file organizes and creates points using the pcta trial data
lines.geojson contains a line associated with a project
projects.geojson contains additional information for each project
The final result will be point keys with hours per mile per year values
"""
"""
Example put_item:

dynamodb.put_item(
    TableName='points', 
    Item={
        'point':{'S':'test'},
        '2022':{'N':'500'}
    }
)
"""


import boto3
import json
import math
from datetime import datetime
from botocore.exceptions import ClientError
date_format = "%m/%d/%Y"


dynamodb = boto3.client('dynamodb')

def close(p1, p2, tolerance=1):
    return math.dist(p1, p2) <= tolerance

# Function to combine lines with tolerance
def combine_lines(lines):
    familiar_points = []
    combined = []
    for line in lines:
        newline = []
        lastpoint = None
        for point in line:
            is_familiar = False
            for checkpoint in familiar_points:
                if close(point, checkpoint):
                    is_familiar = True
                    lastpoint = checkpoint
            if is_familiar:
                newline.append(point)
                if len(newline) > 1:
                    combined.append(newline)
                newline = []
            else:
                if len(newline) == 0 and lastpoint != None:
                    newline.append(lastpoint)
                newline.append(point)
                familiar_points.append(point)
                lastpoint = point
        if len(newline) > 1:
            combined.append(newline)
    return (combined,familiar_points)


#since the values are meters, its easy to get length mathematically
def get_distance(array_3d):
    total_distance = 0
    
    for line in array_3d:
        # Check if there are at least two points in the current second-order list (line)
        if len(line) < 2:
            continue  # Skip this line if it doesn't contain at least two points

        line_distance = 0
        # Loop through consecutive points in the line
        for i in range(len(line) - 1):
            # Get coordinates of the two points (in meters, EPSG:3857)
            x1, y1 = line[i]
            x2, y2 = line[i + 1]
            
            # calculate distance
            distance_in_meters = math.sqrt((x2 - x1)**2 + (y2 - y1)**2)
            
            #convert distance from meters to miles
            distance_in_miles = distance_in_meters / 1609.344
            line_distance += distance_in_miles
        
        # Add this line's distance to the total distance
        total_distance += line_distance
    
    return total_distance

def update_point(point, value, year):
    try:
        # Define stuff
        key_condition = {
            'point': {'S': point}
        }

        # create update expresion
        update_expression = "SET "
        expression_attribute_values = {}

        # check if year in item
        update_expression += f"#year = if_not_exists(#year, :zero) + :num"
        expression_attribute_values = {
            ':num': {'N': str(value)},  # The number to be added
            ':zero': {'N': '0'},         # Default value if the year doesn't exist
        }

        # Map the placeholders to actual attribute names (in case of reserved words)
        expression_attribute_names = {
            '#year': str(year),  # Dynamic year name (e.g., "2025")
        }

        # Perform the update operation
        response = dynamodb.update_item(
            TableName="points",
            Key=key_condition,
            UpdateExpression=update_expression,
            ExpressionAttributeValues=expression_attribute_values,
            ExpressionAttributeNames=expression_attribute_names,
            ReturnValues="ALL_NEW"  # Return the updated item
        )
        
    except ClientError as e:
        print(f"Error updating item: {e}")
        return None


def create_points():
    #for each line:
    with open("key_files/ids.json") as f:
        ids = json.load(f)
    counter = 1
    lengthids = len(ids)
    pointslist = []
    for id in ids:
        #get project project
        project = dynamodb.get_item(
            TableName="projects",
            Key={
                'ID': {'S': id}
            },
            AttributesToGet=["date","hours"]
        )
        #continue if empty project
        if not project['Item']:
            continue
        #get time (total hours by all people) and discard if no time spent
        time = float(project['Item']['hours']['N'])
        if time == 0:
            continue
        #get line
        line = dynamodb.get_item(
            TableName="lines",
            Key={
                'ID': {'S': id}
            },
            AttributesToGet=["points"]
        )
        #continue if empty line
        if not line['Item']:
            continue
        # get points
        point_data = json.loads(line['Item']['points']['S'])

        #random bug fixing stuff (rather than fixing the dataset...)
        #remove empty entries in lines
        point_data = [element for element in point_data if len(element) > 0]
        # fix points in extra array for those weird cases
        depth = lambda L: isinstance(L, list) and max(map(depth, L))+1
        if depth(point_data) == 4:
            point_data = point_data[0]

        # lines is for distance calculation, points is just all the points involved
        lines,points = combine_lines(point_data)
        #get attributes from point into variables
        date = project['Item']['date']['S']
        length = get_distance(lines)
        if length == 0:
            continue
        date1,date2 = date.split(" - ")
        date2=datetime.strptime(date2, date_format)
        year=date2.year
        #get hours per mile
        hourspermile = time/length
        """#print stuff
        print(len(points))
        print(year)
        print(id)
        print(length)
        print(time)
        print(hourspermile)
        print()"""
        #print progress
        print(str(counter)+"/"+str(lengthids))
        #for each point in points section:
        for point in points:
            #creates more similar points
            roundedpoint = [int(round(var, 0)) for var in point]
            if roundedpoint not in pointslist:
                pointslist.append(roundedpoint)
            update_point(json.dumps(roundedpoint),hourspermile,year)
        counter = counter + 1
    with open("key_files/points.json", "w") as f:
        json.dump(pointslist, f)

"""#testing of combine lines method
lines = [
    [[0, 0], [10, 0], [20, 0], [30, 0]],  # Line 1: A-B-C-D
    [[0, 0], [10, 0], [20, 0], [30, 0]],  # Line 2: A-B-C-D (duplicate)
    [[10, 0], [20, 0], [30, 0], [40, 0]],  # Line 3: B-C-D-E (overlap at the end with Line 1)
    [[40, 0], [50, 0]],  # Line 4: E-F (non-overlapping)
    [[0, 0], [10, 0], [20, 0], [30, 0], [40, 0]],  # Line 5: A-B-C-D-E (overlap at the start with Line 3)
    [[60, 0], [70, 0]]  # Line 6: G-H (non-overlapping)
]

condensed_lines = combine_lines(lines)
print(condensed_lines)"""