import json
import joblib
import pickle
import boto3
from threading import Thread
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.model_selection import train_test_split
from sklearn.linear_model import LinearRegression
from sklearn.preprocessing import PolynomialFeatures

dynamodb = boto3.client('dynamodb')

def download_points():
    with open(f"key_files/1000point.json") as f:
        points_list = json.load(f)
    length = len(points_list)
    data = []
    for n, point_array in enumerate(points_list):
        point_str = str(point_array)
        point = dynamodb.get_item(
            TableName="points",
            Key={
                'point': {'S': point_str}
            },
            AttributesToGet=["avg", "traffic", "weather", "slope"]
        )
        point_dict = {"point":point_str}
        attributes = point['Item']
        point_dict["avg"] = attributes['avg']['N']
        point_dict["traffic"] = attributes['traffic']['S']
        point_dict["weather"] = attributes['weather']['S']
        point_dict["slope"] = attributes['slope']['S']
        data.append(point_dict)

        print(f" {n}/{length}\r", end="")
    with open(f"correlated/1000point.json", 'w') as f:
        points_list = json.dump(data,f)

def regression_test(main1, main2, title, additional = None):
    with open(f"correlated/1000point.json") as f:
        points = json.load(f)
    
    ind = {"weather":[],"slope":[],"traffic":[]}
    dep = []
    for point in points:
        ind["weather"].append(float(point["weather"]))
        ind["slope"].append(float(point["slope"]))
        ind["traffic"].append(float(point["traffic"]))
        dep.append(float(point["avg"]))
    #assign data
    X = pd.DataFrame(ind)
    y = pd.Series(dep)

    #filter variables
    if additional == None:
        X = X[[main1, main2]]
    else: 
        X = X[[main1, main2, additional]] 
    print(X)
    #create test train split 
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.1, random_state=42)

    poly = PolynomialFeatures(degree=1)
    X_train_poly = poly.fit_transform(X_train)
    X_test_poly = poly.transform(X_test)

    # initialize and train model
    model = LinearRegression()

    model.fit(X_train_poly, y_train)

    #make predictions code
    y_pred = model.predict(X_test_poly)

    #graph
    fig = plt.figure(figsize=(10, 7))
    ax = fig.add_subplot(111, projection='3d')

    ax.scatter(X_test[main1], X_test[main2], y_test, color='blue', label='Actual Data', s=1.5)

    x1_range = np.linspace(X_test[main1].min(), X_test[main1].max(), 100)
    x2_range = np.linspace(X_test[main2].min(), X_test[main2].max(), 100)
    x1, x2 = np.meshgrid(x1_range, x2_range)

    if additional != None:
        # use average additional
        additional_val = np.median(X_test['weather'])
    
    # generate predictions
    if additional != None:
        grid_points = np.c_[x1.ravel(), x2.ravel(), np.full_like(x1.ravel(), additional_val)]
    else:
        grid_points = np.c_[x1.ravel(), x2.ravel()]
    grid_points_poly = poly.transform(grid_points)  # Apply polynomial feature transformation to the grid
    z = model.predict(grid_points_poly).reshape(x1.shape)

    ax.plot_surface(x1, x2, z, color='red', alpha=0.5, rstride=100, cstride=100)

    ax.set_xlabel(main1.capitalize())
    ax.set_ylabel(main2.capitalize())
    ax.set_zlabel('Work')
    ax.set_title(title)

    plt.show()

def regression():
    with open(f"correlated/1000point.json") as f:
        points = json.load(f)
    
    ind = {"weather":[],"slope":[],"traffic":[]}
    dep = []
    for point in points:
        ind["weather"].append(float(point["weather"]))
        ind["slope"].append(float(point["slope"]))
        ind["traffic"].append(float(point["traffic"]))
        dep.append(float(point["avg"]))
    #assign data
    X = pd.DataFrame(ind)
    y = pd.Series(dep)

    #filter variables
    X = X[['traffic', 'slope', 'weather']] 
    print(X)

    poly = PolynomialFeatures(degree=1)
    X_poly = poly.fit_transform(X)

    # initialize and train model
    model = LinearRegression()

    model.fit(X_poly, y)


    #graph
    fig = plt.figure(figsize=(10, 7))
    ax = fig.add_subplot(111, projection='3d')

    ax.scatter(X['traffic'], X['slope'], y, color='blue', label='Actual Data', s=0.1)

    x1_range = np.linspace(X['traffic'].min(), X['traffic'].max(), 100)
    x2_range = np.linspace(X['slope'].min(), X['slope'].max(), 100)
    x1, x2 = np.meshgrid(x1_range, x2_range)

    # just use average weather
    weather_val = np.median(X['weather'])
    
    # generate predictions
    grid_points = np.c_[x1.ravel(), x2.ravel(), np.full_like(x1.ravel(), weather_val)]
    grid_points_poly = poly.transform(grid_points)  # Apply polynomial feature transformation to the grid
    z = model.predict(grid_points_poly).reshape(x1.shape)

    ax.plot_surface(x1, x2, z, color='red', alpha=0.5, rstride=100, cstride=100)

    ax.set_xlabel('Traffic')
    ax.set_ylabel('Slope')
    ax.set_zlabel('Work')
    ax.set_title('Multiple Linear Regression Best Fit Line (3D)')

    plt.show()
#regression_test('weather', 'traffic', 'Average Yearly Maintenance Data by Weather and Traffic')
#download_points()

def save_model():
    with open(f"correlated/1000point.json") as f:
        points = json.load(f)
    
    ind = {"weather":[],"slope":[],"traffic":[]}
    dep = []
    for point in points:
        ind["weather"].append(float(point["weather"]))
        ind["slope"].append(float(point["slope"]))
        ind["traffic"].append(float(point["traffic"]))
        dep.append(float(point["avg"]))
    #assign data
    X = pd.DataFrame(ind)
    y = pd.Series(dep)


    poly = PolynomialFeatures(degree=1)
    X_poly = poly.fit_transform(X)

    # initialize and train model
    model = LinearRegression()

    model.fit(X_poly, y)

    joblib.dump(poly, 'poly.pkl')

    with open('model.pkl', 'wb') as file:
        pickle.dump(model, file)

    print("model saved")

#save_model()